#
# Math::Quaternion.pm
#
#
#

package Math::Quaternion;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION @CONFIG);
use Carp;


require Exporter;
require DynaLoader;

@ISA = qw(Exporter DynaLoader);
@EXPORT = qw();
@EXPORT_OK = qw(qtrn qtrn_vrotate);

$VERSION = '1.0';

bootstrap Math::Quaternion $VERSION;

use overload
       '+' => '_add',
       '-' => '_sub',
       '*' => '_mul',
       '/' => '_div',
       '~' => '_conjugate',
     'neg' => '_negate',
      '""' => '_stringify',
     'abs' => '_abs',
      '==' => '_equal',
      '!=' => '_not_equal',
       '<' => '_less_than',
      '<=' => '_less_equal',
       '>' => '_greater_than',
      '>=' => '_greater_equal',
     '<=>' => '_spaceship',
       '=' => '_clone',
'nomethod' => '_op_undefined',
'fallback' =>   undef;

my $display = "%g + %gi + %gj + %gk";

#
# A bow to Math::Complex.
#
sub make { &new }

#
# qtrn
#
# Creates a quaternion from a (re, ai, bj, ck) tuple.  Easier than
# typing Math::Quaternion->new(re, ai, bj, ck).
#
sub qtrn
{
	my $r = shift || 0.0;
	my $a = shift || 0.0;
	my $b = shift || 0.0;
	my $c = shift || 0.0;
	return Math::Quaternion->new($r, $a, $b, $c);
}

#
# qtrn_vrotate
#
# Creates a rotation quaternion from an angle and vector
# coordinates.
#
sub qtrn_vrotate
{
	my $angle = shift || 0.0;
	my $b = shift || 0.0;
	my $c = shift || 0.0;
	my $d = shift || 0.0;

	return Math::Quaternion->new_vrotate($angle, $b, $c, $d);
}

#
# ->Re()
#
# Return the real part of the quaternion, in either a scalar
# context or in a list context (a one-element list).
#
sub Re
{
	my($q) = shift;
	my(@r) = $q->parts();
	return wantarray? @r[0..0]: $r[0];
}

#
# ->Im()
#
# Return the imaginary parts of the quaternion, always as a list.
#
sub Im
{
	my($q) = shift;
	my(@i) = $q->parts();
	return @i[1..3];
}

#
# ->_abs()
#
sub _abs
{
	my $self = shift;
	return $self->norm();
}

#
# ->_stringify
#
# Display the quaternion in the re + i + j + k format.
#
sub _stringify
{
	my ($q) = shift;
	return sprintf($display, $q->parts());
}

#
#
#
sub _op_undefined
{
	my($a, $b, $c, $op) = @_;

	carp "Operation '$op' is not defined for objects of class $a";
	return undef;
}

1;
__END__

=head1 NAME

Math::Quaternion - Quaternions and their arithmetic operators.

=head1 SYNOPSIS

	use Math::Quaternion qw(qtrn qtrn_vrotate);

	$point_a = qtrn(0, 1, 1, 1);
	$y_rot = qtrn_vrotate($half_radian, 0, 1, 0);
	$point_b = $point_a->rotate($y_rot);

=head1 DESCRIPTION

This package lets you create and manipulate quaternions in a fairly
natural way.

Quaternions are an extension of complex numbers. They were invented by
William Rowan Hamilton as a means of representing rotations upon vectors
in space, much as complex numbers can represent rotations in the plane.
The number's parts are generally notated by:

B<q> = a + bB<i> + cB<j> + dB<k>

You will note that there are three imaginary parts. This was a sticking
point for Hamilton, since logically if you were to add a dimension to a
number, it would seem that you would only need to add one more imaginary
part to the number, making it a triple. Unfortunately, multiplication
would not work under this method, and it wasn't until ten years later
that that he came up with the notion of using quadruples, and giving up
the commutative rule for multiplying.

Many of the original uses of quaternions have been superceded by other
mathematical techniques, but quaternions are still one of the easiest
methods to use to rotate points in space. This is because rotations can
be combined by simple multiplication.

=head1 CREATION

Quaternions can both represent points and rotations in space. They are
created in separate manners.

=head2 Quaternions as Points

A quaternion represents a point in space by forming it with that point's
coordinates. You use the qtrn() function (found in the EXPORT_OK list),
or the new() or make() member functions, which work identically but
which require more typing:

	$qp = qtrn(0, x, y, z);
	$qp = Math::Quaternion->new(0, x, y, z);
	$qp = Math::Quaternion->make(0, x, y, z);

The member function make() is included as a nod to Math::Complex.

If the last items in the list are zero, you do not have to include them.

	$qp = qtrn(0, 1.618);

is the same as:

	$qp = qtrn(0, 1.618, 0, 0);

=head2 Rotation

To create a quaternion that represents a rotation, you need to provide
the angle of rotation in radians, and the coordinates of the unit vector
that will act as the axis:

	$deg90 = atan2(1,1) * 2;
	$rot_y_axis = qtrn_vrotate($deg90, 0, 1, 0);

If you want to create the rotation quaternion "by hand", the formula is:

	@point = ($x * sin($angle/2), $y * sin($angle/2), $z * sin($angle/2));
	$rvect = qtrn(cos($angle/2), @point);

The actual rotation is performed by the member function rotate().

=head1 ARITHMATIC OPERATIONS

The basic mathematical operators (+, -, *, /) are overloaded for
quaternions, so no functions or members need be called.

=head2 Addition and Subtraction

Addition and subtraction are handled much as you'd expect: the
individual members of each quaternion are added or subtracted to each
other. So,

	2 + 3B<i> + 5B<j> + 7B<k>
     +
 	3 + 5B<i> + 7B<j> + 11B<k>
 	--------------------------------------
 	5 + 8B<i> + 12B<j> + 18B<k>


=head2 Multiplication

Multiplication and division are a trifle more complex. Without getting
into the theory of why it is handled this way (see the references below
for information), multiplication is performed by multiplying each term
by another, following the rules listed below for the imaginary parts:

        |   1    i    j    k
   -----+---------------------
      1 |   1    i    j    k
      i |   i   -1    k   -j
      j |   j   -k   -1    i
      k |   k    j   -i   -1

So, B<k>*B<i> = B<j>, while B<i>*B<k> = -B<j>

As you can see, quaternions are not commutative: S<q1 * q2> does not
equal S<q2 * q1>. But they are associative: S<q1 * (q2 * q3)> will
equal S<(q1 * q2) * q3.>

You may combine rotational quaternions through simple multiplication.
So, if you have complicated series of rotations that you want to apply
to your points, you don't have to perform multiple rotation operations.
You can combine them into a single rotation:

	$deg45 = atan2(1,1);
	$rot45_y_axis = qtrn_vrotate($deg45, 0, 1, 0);
	$rot45_z_axis = qtrn_vrotate($deg45, 0, 0, 1);

	$rot_yz = $rot45_y_axis * $rot45_z_axis;    # Combine the two.
	$qp_rot = $qp->rotate($rot_yz);

=head2 Division

Division is handled by taking the inverse of the divisor, and
multiplying by that. On an individual basis, the inverse of the
imaginary parts are fairly straightforward:

	1/B<i> = -B<i>,
	1/B<j> = -B<j>,
	1/B<k> = -B<k>

=head1 OTHER OPERATORS

=head2 Comparison

The comparison operators (<, <=, ==, =>, >, !=, and <=>) are overloaded
for quaternions. Equality and its opposite are handled as might be
expected: all members of a quaternion must equal the members of another
for them to be considered equal.

The less than and greater than comparisons are handled a little
artifically. The individual members are compared, one at a time, in the
order of real, i, j, and k. If a.re is greater than b.re, a > b. If a.re
is less than b.re, a < b. If the real parts are equal, then the
comparison shifts to the imaginary part B<i>, and so forth.

=head2 Stringification

When printed, a quaternion is shown in its cartesian
form:

	$q1 = qtrn(1.4142, 1.7321, 2.2361, 2.6458)

	print $q1, "\n";

1.4142 + 1.7321B<i> +  2.2361B<j> + 2.6458B<k>

=head2 abs

The absolute value function is considered an operator in perl, and may
be used thusly:

	$val = abs $q1;

$val will hold the norm of $q1.

=head1 MEMBER FUNCTIONS

=head2 new(), make()

The constructors for this class. Member function make() is equivalent to
new(), and is provided solely as a nod to Math::Complex. Components
default to zero if there are no values passed in.

	my $z = Math::Quaternion->new(0, 5, 3);

is identical to

	my $z = Math::Quaternion->new(0, 5, 3, 0);

However, it is easier to invoke qtrn() from EXPORT_OK, and far easier to
create rotation quaternions with qtrn_vrotate().

=head2 Re()

Return the real portion of the quaternion. Return value is a single
number, but it can be in a list context if required. This is to allow
easy switching between Re() and Im().

=head2 Im()

Return the imaginary portion of the quaternion. There are always three
items in the list returned.

=head2 parts()

Return both the real and imaginary parts of the quaternion as a list.

=head2 norm()

Return the absolute value, or norm, of the quaternion.

=head2 rotate()

Rotate a point (quaternion) according to the rotation described in
another quaternion.

	$qp_rotated = $qp->rotate($rot_y_axis);

If you want to apply the rotation quaternion "by hand", the formula is:

	$qp_rotated = (1/$rot_y_axis) * $qp * $rot_y_axis;

=head1 DIAGNOSTICS

Possible errors are using non-numerical values in construction or
calculation, and division by zero. They are handled in a very simple
manner: a warning and the return of an undef.

=head1 REFERENCES

=head2 Perl

On a machine with perl 5 installed, type I<man Math::Complex>

=head2 Books:

B<Mathematics Today>, ed. Lynn Arthur Steen. Vintage Books 1980.
Although "Today" is now two decades ago, the final essay, by Felix E.
Browder and Saunders Mac Lane, explains in five paragraphs what some
authors can't explain in entire chapters. A very well-written collection
of essays.

B<The Book of Numbers>, John H. Conway and Richard K. Guy.
Copernicus (an imprint of Springer-Verlag) 1996.
A survey of the numbers and number systems, including quaternions. A
very enjoyable book.

B<Men of Mathematics>, E. T. Bell.
Fireside (an imprint of Simon and Schuster) 1965.
Although some of the chapters have dubious sources - Bell apparently was
not always careful in checking - the one on Hamilton seems to be true to
his life.

B<A History of Mathematics> Second Edition, Carl B Boyer and Uta
Merzbach. John Wiley & Sons 1991.
A history that covers both the mathematical concepts and the people who
produced them.

=head2 Web and ftp:

I<http://www.treasure-troves.com/math/Quaternion.html>

I<http://http.cs.berkeley.edu/~laura/cs184/quat/quaternion.html>
Computer science course notes on quaternions by Laura Downs.

I<http://www-groups.dcs.st-and.ac.uk/~history/Mathematicians/Hamilton.html>
Biographical information on Hamilton.

Henry Baker's various documents (his own and others that he has posted)
are an incredible resource. See:

I<ftp://ftp.netcom.com/pub/hb/hbaker/home.html>

But in particular:

I<ftp://ftp.netcom.com/pub/hb/hbaker/quaternion>
The quaternion home directory.

I<ftp://ftp.netcom.com/pub/hb/hbaker/quaternion/tait/Encyc-Brit.ps.gz.>
The entry on quaternions in the Ninth Edition of the Encyclopedia
Britannica, by P. G. Tait

I<ftp://ftp.netcom.com/pub/hb/hbaker/quaternion/cmu-cs-82-150.ps.gz>
"Quaternions in Computer Vision and Robotics", by Edward Pervin and Jon A. Web.

I<ftp://ftp.netcom.com/pub/hb/hbaker/quaternion/Silberstein-Relativity.ps.gz>
A derivation of relativity theory by Silberstein, using quaternions.

=head1 ACKNOWLEDGEMENTS

Steffen Beyer may be appalled to learn that my sloppy code was created
after studying his code for Bit::Vector, so I won't mention that.

Raphael Manfredi and Jarkko Hietaniemi may or may not be pleased to know
that I used some of the framework in their Math::Complex package for the
methods and functions here, and that the function qtrn() is stolen
directly from their Math::Complex::cmplx() function, with modifications
of course.

Dean Roehrich's Extension-Building Cookbook A in
I<www.perl.com/CPAN//authors/Dean_Roehrich> was extremely helpful in
creating my typemap. In fact, I only made one change to the code that he
presented.

And to all of the people who post to comp.lang.perl's subgroups - thank you.

=head1 AUTHOR

John M. Gamble may be reached at jgamble@ripco.com

=cut
